--Copyright 2012 Eric Mehl (eric@emdesignviz.com)
--Version: 1.0 rc1
--First release: July 5, 2012
--Last modified: July 5, 2012

--Warning: Use at your own risk!  I try hard to make the scripts as bug free as possible, but I can't be held responsible for
--any problems resulting from their use.  
--Use special caution with mixing alpha/beta/release candidate scripts with production environments.

struct emCamPlus (updateOffsetRegion,updateCamera,camPlusRollout,offsetRollout,isOffsetOpen,iniFileName,dockState,curCam)

debug = true

if debug==true or emCamPlus == undefined then (
	global emCamPlus = emCamPlus()
	emCamPlus.isOffsetOpen = false
)

function snapVal val places =
(
	mult = pow 10 places
	(((((val as float) * mult) as integer) as float) / mult)
)

--this function is called every time something changes that requires the region to be updated.
--it needs to be accessible from callbacks, so it is part of the emCamPlus struct
function updateRegion =
(
	if superclassof emCamPlus.curCam != camera then (return 0)
	
	vpHeight = emCamPlus.camPlusRollout.newHeight as float
	vpWidth = emCamPlus.camPlusRollout.newWidth as float
	
	yoff = abs emCamPlus.curCam.sl_yoffset
	xoff = abs emCamPlus.curCam.sl_xoffset
		
	renderHeight = (1.0 + ((yoff / 100.0) * 2.0)) * vpHeight
	renderWidth = (1.0 + ((xoff / 100.0) * 2.0)) * vpWidth
	
	yoff_norm = emCamPlus.curCam.sl_yoffset / 100.0
	xoff_norm = emCamPlus.curCam.sl_xoffset / 100.0
	
	--first, adjust the fov for the xoffset
	tempFov = 2.0 * atan ((renderWidth / 2.0) / emCamPlus.camPlusRollout.focalDist)
	--tempFov = emCamPlus.camPlusRollout.oldFov * (renderWidth / vpWidth)
	
	emCamPlus.curCam.fov = tempFov
	wVirtual = 2.0 * emCamPlus.camPlusRollout.planeDist * tan (tempFov / 2)
	wTrue = wVirtual * (vpWidth / renderWidth)
	
	--calculate the y offset masking plane	
	emCamPlus.curCam.fovType = 2
	vfov = cameraFOV.FOVtoCurFOV emCamPlus.curCam tempFov
	hVirtual = 2 * emCamPlus.camPlusRollout.planeDist * tan (vfov / 2)
	hTrue = hVirtual * (vpHeight / renderheight)
	
	if debug then (print ("tempFOV=" + (tempFov as string) + " wVirtual=" + (wVirtual as string) + " hVirtual=" + (hVirtual as string) + " vfov=" + (vfov as string) + " oldfov=" + (emCamPlus.camPlusRollout.oldFov as string)))
	
	--set up the y plane mask
	emCamPlus.camPlusRollout.yPlane.length = hVirtual - hTrue
	emCamPlus.camPlusRollout.yPlane.width = 2.0 * emCamPlus.camPlusRollout.planeDist * tan (tempFov / 2)
	ymult = 1.0
	if yoff_norm < 0 then (ymult = -1)
	ypos = (hTrue / 2) * ymult
	xpos = 0.0
	zpos = emCamPlus.camPlusRollout.planeDist
	pt = transMatrix (point3 xpos ypos -zpos)
	pt = pt * emCamPlus.curCam.transform
	emCamPlus.camPlusRollout.yPlane.transform = pt
			
	--set up the x plane mask
	emCamPlus.camPlusRollout.xPlane.length = hVirtual
	emCamPlus.camPlusRollout.xPlane.width = wVirtual - wTrue
	ypos = 0.0
	xmult = -1.0
	if xoff_norm < 0 then (xmult = 1.0)
	xpos = (wTrue / 2.0) * xmult
	pt = transMatrix (point3 xpos ypos -zpos)
	pt = pt * emCamPlus.curCam.transform
	emCamPlus.camPlusRollout.xPlane.transform = pt
)

emCamPlus.updateOffsetRegion = updateRegion

function updateCam =
(
	if superclassof emCamPlus.curCam != camera then (return 0)
	
	--first, adjust the fov for the xoffset
	tempFov = 2.0 * atan ((renderWidth / 2.0) / emCamPlus.camPlusRollout.focalDist)
		
	--find the difference between what the user has currently selected and what it used to be = user's desired change
	--then add that difference to the oldFov to get the new "base" fov, before offset is applied
	if emCamPlus.curCam.fov != tempFov then (
		emCamPlus.camPlusRollout.oldFov = emCamPlus.camPlusRollout.oldFov + emCamPlus.curCam.fov - tempFov
	)
	emCamPlus.updateOffsetRegion()
)
emCamPlus.updateCamera = updateCam

rollout emCamOffset "Maxwell Camera Offset" width:304 height:92
(
	label lbl4 "X Offset" pos:[16,12] width:48 height:16
	label lbl5 "Y Offset" pos:[16,44] width:48 height:16
	slider sldXOffset "" pos:[64,4] width:176 height:25 range:[-100,100,0]
	slider sldYOffset "" pos:[64,32] width:176 height:25 range:[-100,100,0]
	spinner spnXOffset "" pos:[240,12] width:56 height:16 range:[-100,100,0] type:#float scale:0.1
	spinner spnYOffset "" pos:[240,44] width:56 height:16 range:[-100,100,0] type:#float scale:0.1
	button btnLevelCamera "Level Camera" pos:[16,64] width:280 height:24 toolTip:"Reset Camera Target Height to Match Camera Height"
		
	function updateUI =
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		--if the camera is just being created, this function may be called before Maxwell applies its camera info
		--so test to make sure the x / y offset variables are there or else reset to zero
		if isProperty emCamPlus.curCam #sl_xoffset then (
			emCamPlus.offsetRollout.spnXOffset.value = emCamPlus.curCam.sl_xoffset
			emCamPlus.offsetRollout.spnYOffset.value = emCamPlus.curCam.sl_yoffset
		)
		else (
			emCamPlus.offsetRollout.spnXOffset.value = 0
			emCamPlus.offsetRollout.spnYOffset.value = 0
		)
		sldXOffset.value = spnXOffset.value
		sldYOffset.value = spnYOffset.value
	)
	
	----------------------------------------------------------------------------------------------------------------------------
	--OFFSET CALLBACKS
	----------------------------------------------------------------------------------------------------------------------------
	on emCamOffset open do
	(
		emCamPlus.isOffsetOpen = true
		dockState = getINISetting emCamPlus.iniFileName "OffsetDialog" "dockState"
		
		--move the dialog to the old position if it's a floater
		xp = getINISetting emCamPlus.iniFileName "OffsetDialog" "xpos"
		yp = getINISetting emCamPlus.iniFileName "OffsetDialog" "ypos"
		if xp != "" and yp != "" then (
			p2 = point2 (xp as float) (yp as float)
			setDialogPos emCamOffset p2
		)
		cui.RegisterDialogBar emCamOffset minSize:[456, 32] maxSize:[456,32] style:#(#cui_dock_all, #cui_floatable)
		if dockState!="" and dockState != "floating" then (
			cui.DockDialogBar emCamOffset (dockState as name)
		)
		
		updateUI()
	)
	on emCamOffset close do
	(
		emCamPlus.isOffsetOpen = false
		--record dialog state
		if emCamOffset.dialogBar then (
			dockState = (cui.getdockState emCamOffset)
		) 
		else (dockState = "floating")
		setINISetting emCamPlus.iniFileName "OffsetDialog" "dockState" dockState
		pos = getDialogPos emCamOffset
		setINISetting emCamPlus.iniFileName "OffsetDialog" "xpos" (pos.x as string)
		setINISetting emCamPlus.iniFileName "OffsetDialog" "ypos" (pos.y as string)
	)
	
	on sldXOffset changed val do
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		val = snapVal val 1
		spnXOffset.value = val
		emCamPlus.curCam.sl_xoffset = val
	)
	on sldYOffset changed val do
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		val = snapVal val 1
		spnYOffset.value = val
		emCamPlus.curCam.sl_yoffset = val
	)
	on spnXOffset changed val do
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		val = snapVal val 1
		sldXOffset.value = val
		emCamPlus.curCam.sl_xoffset = val
	)
	on spnYOffset changed val do
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		val = snapVal val 1
		sldYOffset.value = val
		emCamPlus.curCam.sl_yoffset = val
	)
	
	on btnLevelCamera pressed do
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		emCamPlus.curCam.target.position.z = emCamPlus.curCam.position.z
	)
)
emCamPlus.offsetRollout = emCamOffset
	
rollout camRollout "Camera+" width:336 height:32
(
	dropdownList ddlCamList "" pos:[8,5] width:230 height:21
	checkbutton btnApplyOffset "S" pos:[246,4] width:22 height:22 toolTip:"Apply Maxwell Camera Offset"
	checkbutton btnLockCamera "L" pos:[276,4] width:22 height:22 toolTip:"Lock Camera Position"
	button btnOffsetRollout "O" pos:[306,4] width:22 height:22 toolTip:"Offset Settings"
	
	----------------------------------------------------------------------------------------------------------------------------
	--LOCAL VARIABLES
	----------------------------------------------------------------------------------------------------------------------------
	local yPlane
	local xPlane
	local oldFov
	local newFov
	local oldWidth
	local newWidth
	local oldHeight
	local newHeight
	local planeDist
	local focalDist
	
	----------------------------------------------------------------------------------------------------------------------------
	--FUNCTIONS
	----------------------------------------------------------------------------------------------------------------------------
	fn applyOffset =
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		--create the masking planes
		yPlane = plane()
		yPlane.lengthSegs = 1
		yPlane.widthSegs = 1
		yPlane.wirecolor = color 0 0 0
		planeDist = 2.0
		--yPlane.xray = true
		yPlane.isFrozen = true
		
		xPlane = plane()
		xPlane.lengthSegs = 1
		xPlane.widthSegs = 1
		xPlane.wirecolor = color 0 0 0
		--xPlane.xray = true
		xPlane.isFrozen = true
		
		--keep track of the current field of view
		oldFov = emCamPlus.curCam.fov
		
		if debug then (print ("oldfov=" + (oldFov as string)))
			
		newFov = oldFov
		focalDist = ((renderwidth as float) / 2.0) / (tan (oldFov / 2.0))
			
		--setup the callback that triggers when the camera parameters change (which will include sl_yoffset / sl_xoffset)
		when parameters emCamPlus.curCam changes id:#offsetCallback do (emCamPlus.updateCamera())
		
		emCamPlus.updateOffsetRegion()
	)
	
	--reset everything and delete the masking planes
	fn removeOffset = 
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		--remove the parameter callback
		deleteAllChangeHandlers id:#offsetCallback
		
		--reset the render height / width
		renderHeight = newHeight
		renderWidth = newWidth
		emCamPlus.curCam.fov = oldFov
		renderSceneDialog.update()
		if isValidNode yPlane then (delete yPlane)
		if isValidNode xPlane then (delete xPlane)
	)
	
	----------------------------------------------------------------------------------------------------------------------------
	--UI / MAX CALLBACK FUNCTIONS
	----------------------------------------------------------------------------------------------------------------------------
	
	function refreshCameraList dbox =
	(
		lst = for obj in objects where (superclassof obj) == camera collect obj.name
		sort lst		
		dbox.ddlCamList.items = lst
		if isValidNode emCamPlus.curCam == false then (
			if lst.count > 0 then (
				emCamPlus.curCam = getNodeByName ddlCamList.items[1]
				ddlCamList.selection = 1
			)
			else (emCamPlus.curCam = undefined)
		)
	)

	function cameraCreated newCam dbox =
	(
		if (superclassof newCam) != camera then (	--only need to worry about cameras
			return 0
		)
			
		emCamPlus.curCam = newCam
		
		--refresh the ui
		refreshCameraList dbox
		
		emCamPlus.camPlusRollout.ddlCamList.selection = (findItem dbox.ddlCamList.items newCam.name)
		
		emCamPlus.camPlusRollout.updateUI()
	)	
	
	function eventRenamed dbox =
	(
		nameSet = callbacks.NotificationParam()
		if ($selection.count ==1) then (
			emCamPlus.camPlusRollout.ddlCamList.selected = nameSet[2]
		)
	)
	
	function eventSelection dbox =
	(
		if debug then (print "event selection")
		
		if $selection.count > 0 then (
			for obj in $selection do (
				if (superclassof obj) == camera then (
					emCamPlus.camPlusRollout.ddlCamList.selection = (findItem dbox.ddlCamList.items obj.name)
					if btnApplyOffset.checked then (removeOffset())
					emCamPlus.curCam = obj
					if btnApplyOffset.checked then (applyOffset())
					emCamPlus.camPlusRollout.updateUI()
				)
			)
		)
	)
	
	function viewportChanged dbox =
	(
		newCam = viewport.getCamera()
		if newCam != undefined and newCam != emCamPlus.curCam then (	--a value of undefined means it's not a camera viewport
			emCamPlus.camPlusRollout.ddlCamList.selection = (findItem emCamPlus.camPlusRollout.ddlCamList.items newCam.name)
			if btnApplyOffset.checked then (removeOffset())
			emCamPlus.curCam = newCam
			if btnApplyOffset.checked then (applyOffset())
			emCamPlus.camPlusRollout.updateUI()
		)
	)
	
	function updateUI =
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		flags = getTransformLockFlags emCamPlus.curCam
		if flags[1] and flags[2] and flags[3] then (
			btnLockCamera.checked = true
		)
		else (btnLockCamera.checked = false)
		
		--update ui for all sub-dialog boxes
		if emCamPlus.isOffsetOpen then (emCamPlus.offsetRollout.updateUI())
	)
	
	----------------------------------------------------------------------------------------------------------------------------
	--EVENT HANDLERS
	----------------------------------------------------------------------------------------------------------------------------
	
	on camRollout open do 
	(
		--add icons to the buttons
		btnApplyOffset.images = #("cameras_16i.bmp","cameras_16a.bmp",3, 1, 1, 1, 1, false, true)
		btnLockCamera.images = #("StatusPanel_16i.bmp","StatusPanel_16a.bmp",35, 1, 3, 2, 4, false, true)
		btnOffsetRollout.images = #("bip_moflow_i.bmp","bip_moflow_i.bmp",10, 9, 9, 9, 9, false, true)
				
		--setup the floating dialog / docking state based on saved ini values
		emCamPlus.iniFileName = ((pathConfig.getDir #plugcfg) + "\camera+.ini")
		dockState = getINISetting emCamPlus.iniFileName "Dialog" "dockState"
		
		--move the dialog to the old position if it's a floater
		xp = getINISetting emCamPlus.iniFileName "Dialog" "xpos"
		yp = getINISetting emCamPlus.iniFileName "Dialog" "ypos"
		if xp != "" and yp != "" then (
			p2 = point2 (xp as float) (yp as float)
			setDialogPos camRollout p2
		)
		cui.RegisterDialogBar camRollout minSize:[336, 32] maxSize:[336,32] style:#(#cui_dock_all, #cui_floatable)
		if dockState != "" and dockState != "floating" then (
			cui.DockDialogBar camRollout (dockState as name)
		)
		
		--update list of cameras
		refreshCameraList camRollout
		
		if ddlCamList.items.count > 0 then (
			newCam = getNodeByName ddlCamList.selected
			if (superclassof newCam) == camera then (
				emCamPlus.curCam = newCam
				updateUI()
			)
		)
		
		--set width and height variables
		oldWidth = renderWidth
		oldHeight = renderHeight
		newWidth = oldWidth
		newHeight = oldHeight
		planeDist = 2.0
		
		--install event callbacks
		callbacks.addScript #sceneNodeAdded "try(camRollout.cameraCreated (callbacks.notificationParam()) (camRollout)) catch(messageBox (getCurrentException()))" id:#camplus
		callbacks.addScript #nodePostDelete "try(camRollout.refreshCameraList camRollout) catch(messageBox (getCurrentException()))" id:#camplus
		callbacks.addScript #systemPostReset "try(camRollout.refreshCameraList camRollout) catch(messageBox (getCurrentException()))" id:#camplus
		callbacks.addScript #filePostOpen "try(camRollout.refreshCameraList camRollout) catch(messageBox (getCurrentException()))" id:#camplus
		callbacks.addScript #nodeRenamed "try(camRollout.eventRenamed camRollout) catch(messageBox (getCurrentException()))" id:#camplus
		--selectionsetChanged is an optional callback - it causes some UI confusion because viewport offsets may not be in sync
		--with the selected camera
		--callbacks.addScript #selectionSetChanged "try(camRollout.eventSelection camRollout) catch(messageBox (getCurrentException()))" id:#camplus
		callbacks.addScript #viewportChange "try(camRollout.viewportChanged (camRollout.ddlCamList)) catch(messageBox (getCurrentException()))" id:#camplus
	)
	
	on camRollout close do
	(
		removeOffset()
		
		--record dialog state
		if camRollout.dialogBar then (dockState = (cui.getdockState camRollout)) 
		else (dockState = "floating")
		setINISetting emCamPlus.iniFileName "Dialog" "dockState" dockState
		pos = getDialogPos emCamPlus.camPlusRollout
		setINISetting emCamPlus.iniFileName "Dialog" "xpos" (pos.x as string)
		setINISetting emCamPlus.iniFileName "Dialog" "ypos" (pos.y as string)
		
		--remove event callbacks
		callbacks.removeScripts id:#camplus
	)
	
	on btnLockCamera changed val do 
	(
		if superclassof emCamPlus.curCam != camera then (return 0)
		
		if val then (
			setTransformLockFlags emCamPlus.curCam #all
		)
		else (
			setTransformLockFlags emCamPlus.curCam #none
		)
	)
	
	on ddlCamList selected val do (
		newCam = getNodeByName ddlCamList.items[val]	
		if (superclassof newCam) == camera then (
			select newCam
			if btnApplyOffset.checked then (removeOffset())
			emCamPlus.curCam = newCam
			if btnApplyOffset.checked then (applyOffset())
			updateUI()
			viewport.setCamera emCamPlus.curCam
		)
	)
	on btnSetView pressed do
	(
		newCam = getNodeByName ddlCamList.selected
		if (superclassof newCam) == camera then (
			viewport.setCamera newCam
		)
	)
	on btnApplyOffset changed val do
	(
		if val then (
			applyOffset()
		)
		else (
			removeOffset()
		)
	)
	on btnOffsetRollout pressed do 
	(
		createDialog emCamPlus.offsetRollout
	)
)

emCamPlus.camPlusRollout = camRollout
createDialog emCamPlus.camPlusRollout
